/* XVCL - the xview classes

   XVCL might one day be a hopefully useful class library
   for xview programming in C++. Just now, it slowly grows
   as I write programs, so many many things are missing.

   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file implements some generic data types.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
*/

#include <string.h>
//  strncmp in list::load
#include <stdio.h>
//  FILE, fopen(), fclose(), fprintf() in the save routines

#include "data.h"

//*********************************************************** list ****

list::list(void)      // create an empty list
{   root = NULL;
    len = 0;
    pos = 0;
    }

int list::init(class listmember * member)
  // if list has been empty when member is added, initialize list,
  // otherwise return zero.
{   if (root) return 0;   // return if list has not been empty
    root = member;    // this member will now be the root
    pos = 1;
    return 1;
    }

int list::before(class listmember * member) const
  // does member appear in the list before root ?
{   class listmember * seek = root->previous();
      // start with one before root
    while(seek)                          // as long as there are members
    {   if (seek == member) return 1;    // is it this one ? yes = return
	seek = seek->previous();         // no: one further back
	}
    return 0;  // member was not among the members tested: return false
    }

void list::remove(class listmember * member)
  // removes member from the list
{   //if (before(member)) pos--;
      // if the member is before root, root's position will be one less
    if (root == member)         // if root points to member
	if (member->next())     //   and is not at the end of the list
	    root = member->next();
	      // root will now point to the next member
	else root = member->previous();
	  // if it is end of the list, root will point to the previous
    len--;                      // list will no be one shorter
//    if (pos > len) pos = len;
      // if member was last element, we set root to the previous member.
      // (see above) incorrect position will be adjusted here.
    member->unlink();
    delete member;                
    }

/*
listmember *list::find(coordinate c)
{   if (!root) return NULL;
    return root->first()->find(c);
    }
*/

void list::save(FILE *f)
  // save this list.
{   listmember *run;

    run = NULL;
    if (root) run = root->first();    // start with first list member
    while(run)              // as long as there are members
    {   run->save(f);       //   save this one
	run = run->next();  //   next member
	}
    }

/*   This needs to be defined by the application itself. It could look like this:
void list::load(FILE *f)
  // load this list. 
{   char s[100];
    listmember *run;

    DELETELIST((*this));
    while (fgets(s, 100, f))
    {   run = new listmember(); 
	run->load(s);
        run->addtolist(*this);
	  // parse data from the buffer string into the list member
        }
    }
*/

// **************************************************** listmember ****

void listmember::addtolist(list &thislist, listmember *aft)
  // insert this object into thislist after member aft, or at end of list
{   prv = nxt = NULL;
    thislist.len++;          // thislist will now be one longer
    if (thislist.init(this)) // initialize list in case it was empty
	return;
    if (!aft) aft = thislist.root->last();
      // append at end of list if aft not specified
    nxt = aft->nxt;   // put new member in pointer chain
    aft->nxt = this;
    prv = aft;
    if (nxt) nxt->prv = this;
    }

listmember::listmember(list &thislist, listmember *aft)
  // create a new member in thislist after member aft, or at end of list
{   addtolist(thislist, aft);
    }

listmember::listmember(void)
  // alternate constructor: listmember outside a list
{   prv = nxt = NULL;
    }

//listmember::~listmember(void)
  // destruct this list member (only to be called from list::remove)
void listmember::unlink(void)
{   if (prv) prv->nxt = nxt;
    if (nxt) nxt->prv = prv;
     // let previous and next member point to each other
    }

listmember * listmember::first(void) const
  // return the first list member:
{   const listmember * seek = this;      // start here,
    while (seek->prv) seek = seek->prv;
      // and go backwards as long as possible
    return (listmember *) seek;
      // cast is necessary to make pointer not const
    }

listmember * listmember::last(void) const
  // return the last member in the list:
{   const listmember * seek = this;      // start here,
    while (seek->nxt) seek = seek->nxt;
      // and go forward as long as possible
    return (listmember *) seek;
      // cast is necessary to make pointer not const
    }

/*
listmember * listmember::find(coordinate c) const
{   if (pos == c) return (listmember *) this; // make unconst
    if (!nxt) return NULL;
    return nxt->find(c);
    }
*/

//*********************************************** coordinate **********

coordinate coordinate::operator + (const coordinate &c)
{   coordinate r;
    r.set(x+c.x, y+c.y);
    return r;
    }

int coordinate::operator - (coordinate &c)
{   int dx, dy;

    dx = x - c.x;
    if (dx < 0) dx = -dx;
    dy = y - c.y;
    if (dy < 0) dy = -dy;

    return (dx > dy) ? dx : dy;
    }

int coordinate::operator == (coordinate &c) const
{   return (x == c.x) && (y == c.y);
    }

